function textBounds = textScroll(w, text, scrollSpeed, format, scrollRect, centerLastLines, enableTextAAS)
% TEXTSCROLL - Make a text scroll from the given text cell array
% This function allows user to create and play text scrollings with minimum
% effort
% Usage:
% --------------------------------------------------
%   textScroll(text, scrollSpeed, format, startY, endY)
%   Input:
%       - w: On screen window handle, acquired from Screen('OpenWindow')
%       - srcRect: the rect for the entire screen, ditto
%       - text: an X by 4 cell array (where X is the number of lines)
%               the first column contain strings for each line, and the
%               second the font name, third the size, fourth the spacing
%               in pixels between the current line and the next line
%             1      2        3        4          5              6
%          string  fontname fontsize vertspacing  indentation    color
%          A line could also be a texture. In this case, the columns
%          contains different parameters
%             1                 2       3       4           5       
%          texture_handle    width    height   vertspacing  indentation
%       - scrollSpeed: how fast the text scrolls, in pixels per sec;
%                      can be followed by a second element which specifies
%                      The time (in secs)it takes from the start until you see
%                       nothing moving on the screen ( i.e. the last line
%                      disappearing outside the scrollRect or a specified
%                      line reaches the center, if centerLastLines is given)
%                      In this case, the first element is 
%                      overriden. This feature enables you to sync your
%                      text scroll with the background music making the
%                      scroll finish precisely when the music is finished
%                       
%       - format: The format of the text, can be 'centered', 'left',
%                 'right'
%       - scrollRect: the 1x4 RECT which defines the boundaries for the
%                     entire scroll
%       - centerLastLines: whether the final lines shall be stopped when
%                          they reach the center of the screen
%                          [] or 0 to disable this feature (default) and
%                          any number N to indicate the last N lines should
%                          be centered; if specified as a 1x2 array [N K],
%                          it means: stop the last N lines when they reach
%                          the center with the last Kth lines exactly
%                          centered. Note that K starts from 0 (the last
%                          line)
%       - enableTextAAS: enables text anti-aliasing or not, [] for default,
%                        see Screen('Preference', 'TextAntiAliasing') for
%                        meanings of numbers.
%   Output:
%       - NONE
% --------------------------------------------------
% ** This function is part of PsychoGame Utility **
% ** Stellari Studio, 2009                       **
% --------------------------------------------------

% Update History
% --------------
% Date                  Modifier                Modifications Made
% -------               ------------            ------------------
% 10/05/09              M. Zhang                Wrote it
% -------               ------------            ------------------
% --------------

%% TODO:
% Add the centerLastLines support

while KbCheck
end

Screen('Preference', 'TextRenderer', 1);
if exist('enableTextAAS', 'var') && ~isempty(enableTextAAS)
    oldTextAASFlag = Screen('Preference', 'TextAntiAliasing', enableTextAAS);
end

if ~exist('centerLastLines','var') || isempty(centerLastLines)
    centerLastLines = 0;
end
currTopPosY = scrollRect(4);    % The top positions of the scroll, starting from the bottom

[nLines nCols] = size(text);         % Get the number of lines
if nCols~= 6
    error('textScroll(): 6 Columns required');
end

TEXT_STRINGS = text(:,1);
TEXT_FONTS = text(:,2);
TEXT_SIZES = text(:,3);
TEXT_LINESPACE = text(:,4);
TEXT_INDENT = text(:,5);
TEXT_COLORS = text(:,6);

textBounds = zeros(nLines, 4);

for i = 1:nLines
    if ischar(TEXT_STRINGS{i})
        Screen('TextFont', w, TEXT_FONTS{i});  % Set the Font and Size
        Screen('TextSize', w, TEXT_SIZES{i});  % so that the program can determine textBounds
        [thisTextBounds] = Screen('TextBounds', w, TEXT_STRINGS{i}, 0, 0);  % Get the text bounds
    else    % i.e. texture handle
        thisTextBounds = [0 0 TEXT_FONTS{i} TEXT_SIZES{i}];
    end
    
    if i == 1
        thisOffsetY = 0;
    else
        lastTextBounds = textBounds(i-1, :);      % Get the last Text Bounds
        
        % Find the bottom y of the last TextBounds and add the spacing to it
        % to get the top y of this text bounds
        thisOffsetY = lastTextBounds(4) + TEXT_LINESPACE{i-1};
    end
    
    % Add the y offset to this Text bounds
    thisTextBounds([2,4]) = thisTextBounds([2,4]) + thisOffsetY;  
    
    switch format   % Find the left x position according to the format
        case 'center'
            thisOffsetX = (scrollRect(3) - thisTextBounds(3))./2; 
        case 'left'
            thisOffsetX = 0;
        case 'right'
            thisOffsetX = scrollRect(3) - thisTextBounds(3);
        otherwise
            error('textScroll(): Unrecognizable format.');
    end
    thisOffsetX = thisOffsetX + TEXT_INDENT{i};
    
    % Update the X values
    thisTextBounds([1,3]) = thisTextBounds([1,3]) + thisOffsetX;  
    
    textBounds(i,:) = thisTextBounds; % Put this bounds in the bounds array
end

%% Process the final lines
if ~isempty(centerLastLines) && centerLastLines(1) ~= 0
    if length(centerLastLines) == 1
        centerLineNum = floor(centerLastLines./2); % Could be 0(the last line)
                                                   % or positive numbers
        if mod(centerLastLines, 1) == 1
            centerLinePosY = (scrollRect(4) + scrollRect(2))./2 - ...    % the center of scroll rect
                            (textBounds(end-centerLineNum, 4) - textBounds(end-centerLineNum, 2))./2;    
                        % half the height of the 'centered line'
        else  % i.e. even numbers
            centerLinePosY = (scrollRect(4) + scrollRect(2))./2 - ...
                            TEXT_LINESPACE{end-centerLineNum}./2;  % half the spacing
        end
    else
        centerLineNum = centerLastLines(2);
        centerLinePosY = (scrollRect(4) + scrollRect(2))./2 - ...    % the center of scroll rect
                            (textBounds(end-centerLineNum, 4) - textBounds(end-centerLineNum, 2))./2;    
                        % half the height of the 'centered line'
    end            
end
%% Process the scrolling time
if length(scrollSpeed) ~= 1
    if exist('centerLinePosY', 'var')   % If the centerLastLines feature enabled
        startY = textBounds(end - centerLineNum, 2) + currTopPosY;    % The line start from this position
        endY = centerLinePosY;         % and end up in this position
        
        % It could happen that the centered lines already stop but the 
        % lines above those lines are still moving and visible. 
        % The program have to find whether the final lines stop first or
        % the lines above final lines disappear first.
        startY2 = textBounds(end - centerLastLines(1), 2) + currTopPosY;   % The last line above 'centerLines' 
        endY2 = scrollRect(2) - ...
            (textBounds(end - centerLastLines(1), 4) - ... 
             textBounds(end - centerLastLines(1), 2));       % which disappears when reached the boundary
        scrollSpeed(1) = max(abs(endY - startY), abs(endY2 - startY2))/scrollSpeed(2);
    else
        startY = textBounds(end, 2) + currTopPosY;    % The last line
        endY = scrollRect(2) - ...      
            (textBounds(end, 4) - ...   
             textBounds(end, 2)) ;      % which disappears when reached the boundary
         scrollSpeed(1) = abs(endY - startY)./scrollSpeed(2);
    end
    scrollSpeed(2:end) = [];    % delete redundant elements
end
    
    

    
%% Now Start drawing and scrolling
firstSec = 0;

% Initialize the two position arrays
currTextPos = [textBounds(:,1), textBounds(:, 2)+currTopPosY];
currTextBot = [textBounds(:,3), textBounds(:, 4)+currTopPosY];

currSec = GetSecs;
for frame = 1:10000
%     Screen('FrameRect', w, [255 255 255], [0 1 1023 768]);
    if currTextPos(end-centerLineNum, 2) > centerLinePosY
        % Get the position(relative to the screen origin)
        currTextPos = [textBounds(:,1), textBounds(:, 2)+currTopPosY];
        % Get the bottom position for each string
        % Only used to determine whether a text is visible
        currTextBot = [textBounds(:,3), textBounds(:, 4)+currTopPosY];
    else %  If the 'center line' reaches the center, the last N lines will stop moving
        currTextPos(1:end-centerLastLines(1), :) = ...
            [textBounds(1:end-centerLastLines(1),1), ...
            textBounds(1:end-centerLastLines(1),2)+currTopPosY];
        
        currTextBot(1:end-centerLastLines(1), :) = ...
            [textBounds(1:end-centerLastLines(1),3), ...
            textBounds(1:end-centerLastLines(1),4)+currTopPosY];
    end
    
    if currTextBot(end, 2) < scrollRect(2)  % If everything is out, quit immediately
        break;
    end
    % Select those visible in the screen
    
    visibleTextPosInd = find((currTextPos(:, 2) < scrollRect(4) & ...
                         currTextPos(:, 2) > scrollRect(2)) |...
                        (currTextBot(:, 2) < scrollRect(4) & ...
                         currTextBot(:, 2) > scrollRect(2)));
%     visibleTextPos = currTextPos(visibleTextPosInd);
    
% Draw all the visible strings
start = GetSecs;
    if ~isempty(visibleTextPosInd)
        if size(visibleTextPosInd, 1) > size(visibleTextPosInd, 2)
            visibleTextPosInd = visibleTextPosInd'; % for loop accepts only row vectors
        end
        
        for i = visibleTextPosInd   %(1):visibleTextPosInd(end)
%             TEXT_STRINGS{i};
%             currTextPos(i,1);
%             currTextPos(i,2);
            if ischar(TEXT_STRINGS{i})
                Screen('TextFont', w, TEXT_FONTS{i});
                Screen('TextSize', w, TEXT_SIZES{i});
                Screen('DrawText', w, TEXT_STRINGS{i}, currTextPos(i ,1), currTextPos(i, 2), TEXT_COLORS{i});
            else % i.e. texture handle
                % Draw the picture in the right position
                Screen('DrawTexture', w, TEXT_STRINGS{i}, [], [currTextPos(i, :) currTextBot(i,:)]);
            end
        end
        
    end
    endt = GetSecs;
%     if endt - start > 0.013
%         fprintf('Frame: %d, Time it takes: %.6f\n', frame,endt-start);
%     end
    % Display
    currSec = Screen('Flip', w);    
    if firstSec == 0 % meaning this is the first cycle
        firstSec = currSec; % The first timestamp (beginning)
    end

    % The distance moved since the beginning
    currTopPosY = scrollRect(4)  -  (currSec - firstSec).* scrollSpeed;
    
%     lastSec = GetSecs;
    [keyIsDown, secs, keyCode] = KbCheck;
    if keyCode(KbName('escape'))
        break;
    end
end
    
if exist('enableTextAAS', 'var') && ~isempty(enableTextAAS)
Screen('Preference', 'TextAntiAliasing', oldTextAASFlag);   % Restore the old antialiasing setting
end



